import itertools
import logging
import typing as t

import ase
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from pydantic import Field

from zndraw.base import Extension
from zndraw.utils import update_figure_layout

try:
    from zndraw.analyse import mda  # noqa: F401
except ImportError:
    # mdanalysis is not installed
    pass


def _get_available_properties(atoms: ase.Atoms) -> list[str]:
    available_properties = list(atoms.arrays.keys())
    available_properties += list(atoms.info.keys())
    if atoms.calc is not None:
        available_properties += list(atoms.calc.results.keys())
    return available_properties


log = logging.getLogger(__name__)


def _get_data_from_frames(key, frames: list[ase.Atoms]):
    if frames[0].calc is not None and key in frames[0].calc.results:
        data = np.array([x.calc.results[key] for x in frames])
    elif key in frames[0].arrays:
        data = np.array([x.arrays[key] for x in frames])
    elif key in frames[0].info:
        data = np.array([x.info[key] for x in frames])
    else:
        raise ValueError(f"Property '{key}' not found in atoms")

    return data


class AnaylsisMethod(Extension):
    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        return cls.model_json_schema()


class DihedralAngle(AnaylsisMethod):
    def run(self, vis):
        atoms_lst = list(vis)
        dihedral_angles = []

        if len(vis.selection) != 4:
            raise ValueError("Please select exactly 4 atoms")
        for atoms in atoms_lst:
            dihedral_angles.append(
                atoms.get_dihedrals(indices=[vis.selection], mic=True)[0]
            )
        df = pd.DataFrame(
            {"step": list(range(len(atoms_lst))), "dihedral": dihedral_angles}
        )
        fig = px.line(df, x="step", y="dihedral", render_mode="svg")

        meta_step = np.arange(len(atoms_lst))
        # meta_idx = np.full_like(meta_step, np.nan)

        fig.update_traces(
            customdata=np.stack([meta_step], axis=-1),
        )
        update_figure_layout(fig)

        vis.figures.update({"DihedralAngle": fig})


class Distance(AnaylsisMethod):
    smooth: bool = False
    mic: bool = True

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()

        schema["properties"]["smooth"]["format"] = "checkbox"
        schema["properties"]["mic"]["format"] = "checkbox"

        return schema

    def run(self, vis):
        atoms_lst, ids = list(vis), vis.selection
        distances = {}
        for x in itertools.combinations(ids, 2):
            distances[f"{tuple(x)}"] = []
        for atoms in atoms_lst:
            for x in itertools.combinations(ids, 2):
                distances[f"{tuple(x)}"].append(
                    atoms.get_distance(x[0], x[1], mic=self.mic)
                )

        df = pd.DataFrame({"step": list(range(len(atoms_lst)))} | distances)

        fig = px.line(
            df,
            x="step",
            y=df.columns,
            title="Distance between selected particles",
            render_mode="svg",  # This is important, otherwise openGL will be used
            # and there can/will be issues with three.js
        )
        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )
        meta_step = np.arange(len(atoms_lst))
        # meta_idx = np.full_like(meta_step, np.nan)

        fig.update_traces(
            customdata=np.stack([meta_step], axis=-1),
        )
        update_figure_layout(fig)

        vis.figures.update({"Distance": fig})


class Properties2D(AnaylsisMethod):
    x_data: str
    y_data: str
    color: str
    fix_aspect_ratio: bool = True

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()

        available_properties = _get_available_properties(atoms)

        available_properties += ["step"]
        schema["properties"]["x_data"]["enum"] = available_properties
        schema["properties"]["y_data"]["enum"] = available_properties
        schema["properties"]["color"]["enum"] = available_properties
        schema["properties"]["fix_aspect_ratio"]["format"] = "checkbox"

        return schema

    def run(self, vis):
        atoms_lst = list(vis)
        log.info(f"run {self}")

        if self.x_data == "step":
            x_data = list(range(len(atoms_lst)))
        else:
            x_data = _get_data_from_frames(self.x_data, atoms_lst)

        if self.y_data == "step":
            y_data = list(range(len(atoms_lst)))
        else:
            y_data = _get_data_from_frames(self.y_data, atoms_lst)

        if self.color == "step":
            color = list(range(len(atoms_lst)))
        else:
            color = _get_data_from_frames(self.color, atoms_lst)

        y_data = np.array(y_data).reshape(-1)
        x_data = np.array(x_data).reshape(-1)
        color = np.array(color).reshape(-1)

        df = pd.DataFrame({self.x_data: x_data, self.y_data: y_data, self.color: color})
        fig = px.scatter(df, x=self.x_data, y=self.y_data, color=self.color)
        if self.fix_aspect_ratio:
            fig.update_yaxes(
                scaleanchor="x",
                scaleratio=1,
            )

        meta_step = np.arange(len(atoms_lst))
        # meta_idx = np.full_like(meta_step, np.nan)

        fig.update_traces(
            customdata=np.stack([meta_step], axis=-1),
        )
        update_figure_layout(fig)

        vis.figures.update({"Properties2D": fig})


class ForceCorrelation(AnaylsisMethod):
    """Compute the correlation between two properties for the current frame."""

    x_data: str
    y_data: str

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()

        available_properties = _get_available_properties(atoms)
        schema["properties"]["x_data"]["enum"] = available_properties
        schema["properties"]["y_data"]["enum"] = available_properties

        return schema

    def run(self, vis):
        atoms = vis.atoms
        step = vis.step
        x_data = _get_data_from_frames(self.x_data, [atoms])
        y_data = _get_data_from_frames(self.y_data, [atoms])

        x_data = np.linalg.norm(x_data, axis=-1)
        y_data = np.linalg.norm(y_data, axis=-1)

        vis.log(f"x_data: {x_data.shape}, y_data: {y_data.shape}")

        x_data = x_data.reshape(-1)
        y_data = y_data.reshape(-1)

        meta_step = [None for _ in range(len(x_data))]
        meta_idx = list(range(len(x_data)))

        df = pd.DataFrame(
            {
                self.x_data: x_data,
                self.y_data: y_data,
            }
        )

        fig = px.scatter(df, x=self.x_data, y=self.y_data, render_mode="svg")
        fig.update_traces(customdata=np.stack([meta_step, meta_idx], axis=-1))

        fig.update_layout(
            title=f"Force Correlation for step {step}",
            xaxis_title=self.x_data,
            yaxis_title=self.y_data,
        )
        update_figure_layout(fig)

        vis.figures.update({f"ForceCorrelation-{step}": fig})


class Properties1D(AnaylsisMethod):
    value: str
    smooth: bool = False

    aggregation: t.Literal["mean", "median", "max", ""] = Field(
        "",
        description="For multidimensional data, aggregate over all dimensions, except the first one.",
    )

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()

        available_properties = _get_available_properties(atoms)
        schema["properties"]["value"]["enum"] = available_properties

        return schema

    def run(self, vis):
        vis.log("Downloading data...")
        atoms_lst = list(vis)
        data = _get_data_from_frames(self.value, atoms_lst)

        if data.ndim > 1:
            axis = tuple(range(1, data.ndim))
            if self.aggregation == "mean":
                data = np.mean(data, axis=axis)
            elif self.aggregation == "median":
                data = np.median(data, axis=axis)
            elif self.aggregation == "max":
                data = np.max(data, axis=axis)

        df = pd.DataFrame({"step": list(range(len(atoms_lst))), self.value: data})

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=df["step"],
                y=df[self.value],
                mode="lines+markers",
            )
        )
        # set xlabel to be step
        fig.update_layout(
            xaxis_title="step",
            yaxis_title=self.value,
        )

        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )

        meta_step = np.arange(len(atoms_lst))
        # meta_idx = np.full_like(meta_step, np.nan)

        fig.update_traces(
            customdata=np.stack([meta_step], axis=-1),
        )
        update_figure_layout(fig)

        vis.figures.update({"Properties1D": fig})


analyses: dict[str, t.Type[AnaylsisMethod]] = {
    Properties1D.__name__: Properties1D,
    DihedralAngle.__name__: DihedralAngle,
    Distance.__name__: Distance,
    Properties2D.__name__: Properties2D,
    ForceCorrelation.__name__: ForceCorrelation,
}



import dataclasses
import datetime
import enum
import importlib.metadata
import logging
import typing as t
from collections.abc import MutableSequence

import ase
import numpy as np
import plotly.graph_objects as go
import requests
import socketio.exceptions
import splines
import tqdm
import typing_extensions as tyex
import znjson
import znsocket
import znsocket.exceptions
from redis import Redis

from zndraw.abc import Message
from zndraw.base import Extension
from zndraw.bonds import ASEComputeBonds
from zndraw.config import SETTINGS
from zndraw.converter import ASEConverter, Object3DConverter
from zndraw.draw import Object3D
from zndraw.exceptions import RoomLockedError
from zndraw.figure import Figure, FigureConverter
from zndraw.queue import check_queue
from zndraw.type_defs import (
    ATOMS_LIKE,
    CameraData,
    JupyterConfig,
    RegisterModifier,
    TimeoutConfig,
)
from zndraw.utils import (
    call_with_retry,
    convert_url_to_http,
    emit_with_retry,
    parse_url,
)

log = logging.getLogger(__name__)
__version__ = importlib.metadata.version("zndraw")


class ExtensionType(str, enum.Enum):
    """The type of the extension."""

    MODIFIER = "modifier"
    SELECTION = "selection"
    ANALYSIS = "analysis"


def _check_version_compatibility(server_version: str) -> None:
    if server_version != __version__:
        log.warning(
            f"Server version ({server_version}) and client version ({__version__}) are not the same. "
            "This may lead to unexpected behavior."
        )


@dataclasses.dataclass
class ZnDraw(MutableSequence):
    url: str
    token: str | None = None
    auth_token: str | None = None
    r: Redis | znsocket.Client | None = None

    socket: socketio.Client | None = dataclasses.field(default=None, repr=False)
    timeout: TimeoutConfig = dataclasses.field(
        default_factory=lambda: TimeoutConfig(
            connection=10,
            modifier=0.25,
            between_calls=0.1,
            emit_retries=3,
            call_retries=3,
            connect_retries=3,
        )
    )
    jupyter_config: JupyterConfig = dataclasses.field(
        default_factory=lambda: JupyterConfig(
            width="70%",
            height=600,
        )
    )
    verify: bool | str = True
    convert_nan: bool = False

    max_atoms_per_call: int = dataclasses.field(default=1000, repr=False)
    # number of `ase.Atom` to send per call
    name: str | None = None

    _modifiers: dict[str, RegisterModifier] = dataclasses.field(default_factory=dict)
    _available: bool = True
    _refresh_client: znsocket.Client | None = None

    bond_calculator: ASEComputeBonds | None = dataclasses.field(
        default_factory=ASEComputeBonds, repr=False
    )

    def __post_init__(self):
        if self.socket is None:
            http_session = requests.Session()
            http_session.verify = self.verify
            self.socket = socketio.Client(http_session=http_session)

        self._refresh_client = znsocket.Client.from_url(self.url)
        # TODO: the refresh_client should be able to use the same socket connection!
        if self.r is None:
            self.r = self._refresh_client

        self.url = self.url.replace("http", "ws")
        self.socket.on("connect", self._on_connect)
        self.socket.on("room:log", lambda x: print(x))
        self.socket.on("version", _check_version_compatibility)
        self.socket.on("disconnect", lambda: print("Disconnected from ZnDraw server"))

        self._connect()

        self.socket.start_background_task(check_queue, self)

    def _connect(self):
        for idx in range(self.timeout["connect_retries"] + 1):
            try:
                _url, _path = parse_url(self.url)
                if _path:
                    self.socket.connect(
                        _url,
                        wait_timeout=self.timeout["connection"],
                        socketio_path=f"/{_path}/socket.io",
                    )
                else:
                    self.socket.connect(_url, wait_timeout=self.timeout["connection"])
                break
            except socketio.exceptions.ConnectionError as err:
                log.warning("Connection failed. Retrying...")
                self.socket.sleep(self.timeout["connection"])
                if idx == self.timeout["connect_retries"]:
                    raise socketio.exceptions.ConnectionError(
                        f"Unable to connect to ZnDraw server at '{self.url}'. Is the server running?"
                    ) from err

    def _on_connect(self):
        log.debug("Connected to ZnDraw server")

        emit_with_retry(
            self.socket,
            "join",
            {
                "token": str(self.token),
                "auth_token": self.auth_token,
            },
            retries=self.timeout["emit_retries"],
        )

        # self.socket.sleep(5)  # wait for znsocket to reconnect as well
        registerd_modifiers = list(self._modifiers[x]["cls"] for x in self._modifiers)
        for modifier in registerd_modifiers:
            try:
                self.register_modifier(
                    modifier,
                    public=self._modifiers[modifier.__name__]["public"],
                    run_kwargs=self._modifiers[modifier.__name__]["run_kwargs"],
                )
            except (
                socketio.exceptions.BadNamespaceError
            ) as err:  # /znsocket is not a connected namespace.
                log.error(str(err))
                log.error("Disconnecting. Please restart.")
                self.socket.disconnect()

    def __getitem__(self, index: int | list | slice) -> ase.Atoms | list[ase.Atoms]:
        single_item = isinstance(index, int)
        if single_item:
            index = [index]
        if self.r.exists(f"room:{self.token}:frames"):
            structures = znsocket.List(
                self.r,
                f"room:{self.token}:frames",
                converter=[ASEConverter],
                socket=self._refresh_client,
                max_commands_per_call=100,
            )[index]

        else:
            structures = znsocket.List(
                self.r,
                "room:default:frames",
                converter=[ASEConverter],
                socket=self._refresh_client,
                max_commands_per_call=100,
            )[index]

        if single_item:
            return structures[0]
        return structures

    def __setitem__(
        self,
        index: int | list | slice,
        value: ase.Atoms | list[ase.Atoms],
    ):
        if isinstance(value, list):
            if not all(isinstance(x, ase.Atoms) for x in value):
                raise ValueError("Unable to parse provided data object")
        else:
            if not isinstance(value, ase.Atoms):
                raise ValueError("Unable to parse provided data object")
        if self.locked:
            raise RoomLockedError
        lst = znsocket.List(
            self.r,
            f"room:{self.token}:frames",
            converter=[ASEConverter],
            socket=self._refresh_client,
            max_commands_per_call=100,
            convert_nan=self.convert_nan,
        )
        if not self.r.exists(f"room:{self.token}:frames") and self.r.exists(
            "room:default:frames"
        ):
            default_lst = znsocket.List(
                self.r,
                "room:default:frames",
                socket=self._refresh_client,
                max_commands_per_call=100,
            )
            if not (len(default_lst) == 1 and len(default_lst[0]) == 0):
                # prevent copying empty default room
                default_lst.copy(key=lst.key)

        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))
        if isinstance(index, int):
            index = [index]
            value = [value]

        for i, val in zip(index, value):
            if isinstance(val, ase.Atoms):
                if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                    val.connectivity = self.bond_calculator.get_bonds(val)
            lst[i] = val

    def __len__(self) -> int:
        # TODO: what if the room does not exist yet?
        if not self.r.exists(f"room:{self.token}:frames"):
            return len(
                znsocket.List(
                    self.r,
                    "room:default:frames",
                    socket=self._refresh_client,
                )
            )
        return len(
            znsocket.List(
                self.r,
                f"room:{self.token}:frames",
                socket=self._refresh_client,
            )
        )

    def __delitem__(self, index: int | slice | list[int]):
        if self.locked:
            raise RoomLockedError
        lst = znsocket.List(
            self.r,
            f"room:{self.token}:frames",
            converter=[ASEConverter],
            socket=self._refresh_client,
            max_commands_per_call=100,
        )

        if not self.r.exists(f"room:{self.token}:frames") and self.r.exists(
            "room:default:frames"
        ):
            default_lst = znsocket.List(
                self.r,
                "room:default:frames",
                converter=[ASEConverter],
                socket=self._refresh_client,
                max_commands_per_call=100,
            )
            if not (len(default_lst) == 1 and len(default_lst[0]) == 0):
                # prevent copying empty default room
                default_lst.copy(key=lst.key)

        del lst[index]

    def _repr_html_(self):
        from IPython.display import IFrame

        # TODO: save address and do not replace in post_init
        address = convert_url_to_http(f"{self.url}/token/{self.token}")
        log.info(f"Opening ZnDraw at {address}")
        return IFrame(
            address,
            width=self.jupyter_config["width"],
            height=self.jupyter_config["height"],
        )._repr_html_()

    def insert(self, index: int, value: ase.Atoms):
        if not isinstance(value, ase.Atoms):
            raise ValueError("Unable to parse provided data object")
        if self.locked:
            raise RoomLockedError
        lst = znsocket.List(
            self.r,
            f"room:{self.token}:frames",
            converter=[ASEConverter],
            socket=self._refresh_client,
            convert_nan=self.convert_nan,
        )
        if not self.r.exists(f"room:{self.token}:frames") and self.r.exists(
            "room:default:frames"
        ):
            default_lst = znsocket.List(
                self.r,
                "room:default:frames",
                converter=[ASEConverter],
                socket=self._refresh_client,
            )
            if not (len(default_lst) == 1 and len(default_lst[0]) == 0):
                # prevent copying empty default room
                default_lst.copy(key=lst.key)

        if isinstance(value, ase.Atoms):
            if not hasattr(value, "connectivity") and self.bond_calculator is not None:
                value.connectivity = self.bond_calculator.get_bonds(value)
        lst.insert(index, value)

    def extend(self, values: list[ase.Atoms]):
        if not isinstance(values, list) or not all(
            isinstance(x, ase.Atoms) for x in values
        ):
            raise ValueError("Unable to parse provided data object")
        if self.locked:
            raise RoomLockedError

        if len(values) == 0:
            return

        if not self.r.exists(f"room:{self.token}:frames") and self.r.exists(
            "room:default:frames"
        ):
            default_lst = znsocket.List(
                self.r,
                "room:default:frames",
                converter=[ASEConverter],
                socket=self._refresh_client,
                max_commands_per_call=100,
            )
            if not (len(default_lst) == 1 and len(default_lst[0]) == 0):
                # prevent copying empty default room
                default_lst.copy(key=f"room:{self.token}:frames")

        # enable tbar if more than 10 messages are sent
        # approximated by the size of the first frame
        lst = znsocket.List(
            self.r,
            f"room:{self.token}:frames",
            converter=[ASEConverter],
            socket=self._refresh_client,
            max_commands_per_call=100,
            convert_nan=self.convert_nan,
        )
        # TODO: why is there no copy action here?
        show_tbar = (len(values[0]) * len(values)) > self.max_atoms_per_call
        tbar = tqdm.tqdm(
            values, desc="Sending frames", unit=" frame", disable=not show_tbar
        )

        msg = []
        n_atoms = 0

        for val in tbar:
            if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                val.connectivity = self.bond_calculator.get_bonds(val)

            msg.append(val)
            n_atoms += len(val)

            if n_atoms > self.max_atoms_per_call:
                lst.extend(msg)
                msg = []
                n_atoms = 0
        if len(msg) > 0:
            lst.extend(msg)

    @property
    def selection(self) -> list[int]:
        try:
            return znsocket.Dict(
                self.r, f"room:{self.token}:selection", socket=self._refresh_client
            )["grp-0"]
        except KeyError:
            return []

    @selection.setter
    def selection(self, value: list[int]):
        if not isinstance(value, list):
            raise ValueError("Selection must be a list")
        if not all(isinstance(x, int) for x in value):
            raise ValueError("Selection must be a list of integers")
        if len(value) != len(set(value)):
            raise ValueError("Selection must not contain duplicates")
        if self.locked:
            raise RoomLockedError

        max_index = len(self.atoms)
        if any(x >= max_index for x in value):
            raise IndexError("Selection out of range")
        if any(x < 0 for x in value):
            raise IndexError("Selection must be positive")
        znsocket.Dict(
            self.r, f"room:{self.token}:selection", socket=self._refresh_client
        )["grp-0"] = value

    @property
    def step(self) -> int:
        try:
            return int(
                znsocket.Dict(
                    self.r, f"room:{self.token}:step", socket=self._refresh_client
                )["grp-0"]
            )
        except KeyError:
            return 0

    def log(self, message: str) -> None:
        msg: Message = {
            "time": datetime.datetime.now().isoformat(),
            "msg": message,
            "origin": self.name,
        }
        znsocket.List(
            self.r,
            f"room:{self.token}:chat",
            socket=self._refresh_client,
            max_commands_per_call=100,
            convert_nan=self.convert_nan,
        ).append(msg)

    @property
    def messages(self) -> list[Message]:
        return znsocket.List(
            self.r,
            f"room:{self.token}:chat",
            repr_type="length",
            socket=self._refresh_client,
            max_commands_per_call=100,
        )

    @step.setter
    def step(self, value: int):
        if not isinstance(value, int):
            raise ValueError("Step must be an integer")
        if value < 0:
            raise ValueError("Step must be positive")
        if value >= len(self):
            raise IndexError("Step out of range")
        if self.locked:
            raise RoomLockedError
        # Have only one step per room!
        # shared rooms are rare anyhow and making per-client steps
        # and room hosts is annoying
        # what about the camera?
        # or collect the steps of all clients in a dict
        # and save the host and go from there, also fine and not too much worker.
        znsocket.Dict(self.r, f"room:{self.token}:step", socket=self._refresh_client)[
            "grp-0"
        ] = value

    @property
    def figures(self) -> dict[str, go.Figure | Figure]:
        return znsocket.Dict(
            self.r,
            f"room:{self.token}:figures",
            repr_type="full",
            socket=self._refresh_client,
            converter=[znjson.converter.PlotlyConverter, FigureConverter],
        )

    @figures.setter
    def figures(self, data: dict[str, go.Figure]) -> None:
        """Update the figures on the remote."""
        if self.locked:
            raise RoomLockedError
        figures_dict = znsocket.Dict(
            self.r,
            f"room:{self.token}:figures",
            socket=self._refresh_client,
            converter=[znjson.converter.PlotlyConverter, FigureConverter],
        )
        figures_dict.clear()
        figures_dict.update(data)

    @property
    def atoms(self) -> ase.Atoms:
        try:
            return self[self.step]
        except IndexError:
            return ase.Atoms()

    @atoms.setter
    def atoms(self, atoms: ATOMS_LIKE) -> None:
        self[[self.step]] = [atoms]

    @property
    def points(self) -> np.ndarray:
        # TODO: use znsocket.List inside znsocket.Dict here
        try:
            return np.array(
                znsocket.Dict(
                    self.r, f"room:{self.token}:points", socket=self._refresh_client
                )["grp-0"]
            )
        except KeyError:
            return np.array([])

    @points.setter
    def points(self, points: np.ndarray | list) -> None:
        if self.locked:
            raise RoomLockedError
        if isinstance(points, np.ndarray):
            points = points.tolist()
        znsocket.Dict(self.r, f"room:{self.token}:points", socket=self._refresh_client)[
            "grp-0"
        ] = points

    @property
    def segments(self) -> np.ndarray:
        points = self.points
        if points.shape[0] <= 1:
            return points
        t = np.linspace(0, len(points) - 1, len(points) * 50)
        return splines.CatmullRom(points).evaluate(t)

    @property
    def bookmarks(self) -> dict[int, str]:
        return znsocket.Dict(
            self.r,
            f"room:{self.token}:bookmarks",
            repr_type="full",
            socket=self._refresh_client,
        )

    @bookmarks.setter
    def bookmarks(self, value: dict[int, str]):
        if self.locked:
            raise RoomLockedError
        if not isinstance(value, dict):
            raise ValueError("Bookmarks must be a dictionary")
        if not all(isinstance(x, int) for x in value.keys()):
            raise ValueError("Bookmark keys must be integers")
        if not all(isinstance(x, str) for x in value.values()):
            raise ValueError("Bookmark values must be strings")

        bookmarks = znsocket.Dict(
            self.r, f"room:{self.token}:bookmarks", socket=self._refresh_client
        )
        bookmarks.clear()
        bookmarks.update(value)

    @property
    def camera(self) -> CameraData:
        camera_dct = znsocket.Dict(
            self.r,
            f"room:{self.token}:camera",
            repr_type="full",
            socket=self._refresh_client,
        )
        if "position" not in camera_dct:
            camera_dct["position"] = [5, 5, 5]
        if "target" not in camera_dct:
            camera_dct["target"] = [0, 0, 0]
        return camera_dct

    @camera.setter
    def camera(self, value: CameraData):
        if self.locked:
            raise RoomLockedError
        if set(value.keys()) != {"position", "target"}:
            raise ValueError("Camera must have 'position' and 'target' keys")
        znsocket.Dict(
            self.r, f"room:{self.token}:camera", socket=self._refresh_client
        ).update(value)

    @property
    def geometries(self) -> list[Object3D]:
        return znsocket.List(
            self.r,
            f"room:{self.token}:geometries",
            repr_type="full",
            socket=self._refresh_client,
            converter=[Object3DConverter],
            max_commands_per_call=100,
        )

    @geometries.setter
    def geometries(self, value: list[Object3D]):
        if not all(isinstance(x, Object3D) for x in value):
            raise ValueError("Geometries must be a list of Object3D instances")
        if self.locked:
            raise RoomLockedError
        lst = znsocket.List(
            self.r,
            f"room:{self.token}:geometries",
            socket=self._refresh_client,
            converter=[Object3DConverter],
            max_commands_per_call=100,
        )
        lst.clear()
        lst.extend(value)

    @property
    def config(self) -> znsocket.Dict:
        conf = znsocket.Dict(
            self.r,
            f"room:{self.token}:config",
            repr_type="full",
            socket=self._refresh_client,
        )
        if len(conf) == 0:
            conf_dicts = {}
            for key, value in SETTINGS.items():
                conf_dicts[key] = znsocket.Dict(
                    self.r,
                    f"room:{self.token}:config:{key}",
                    repr_type="full",
                    socket=self._refresh_client,
                )
                if len(conf_dicts[key]) == 0:
                    conf_dicts[key].update(value().model_dump())
            conf.update(conf_dicts)
            # TODO: arrows
        return conf

    @property
    def locked(self) -> bool:
        return call_with_retry(self.socket, "room:lock:get")

    @locked.setter
    def locked(self, value: bool) -> None:
        emit_with_retry(self.socket, "room:lock:set", value)

    def register(
        self,
        cls: t.Type[Extension],
        run_kwargs: dict | None = None,
        public: bool = False,
        variant: ExtensionType = ExtensionType.MODIFIER,
    ):
        """Register an extension class."""
        if run_kwargs is None:
            run_kwargs = {}

        if variant == ExtensionType.MODIFIER:
            # TODO: need to create a custom vis object for the room!
            if public:
                modifier_schema = znsocket.Dict(
                    self.r,
                    "schema:default:modifier",
                    socket=self._refresh_client,
                )
            else:
                modifier_schema = znsocket.Dict(
                    self.r,
                    f"schema:{self.token}:modifier",
                    socket=self._refresh_client,
                )
            # TODO: check if the key exists and if it is different?
            # TODO: also check in the default schema room.
            # TODO: can not register the same modifier twice!
            modifier_schema[cls.__name__] = cls.model_json_schema()

            # TODO: if public!
            if public:
                modifier_registry = znsocket.Dict(
                    self.r,
                    "registry:default:modifier",
                    socket=self._refresh_client,
                )
            else:
                modifier_registry = znsocket.Dict(
                    self.r,
                    f"registry:{self.token}:modifier",
                    socket=self._refresh_client,
                )
            if self.socket.get_sid() not in modifier_registry:
                modifier_registry[self.socket.get_sid()] = [cls.__name__]
            else:
                modifier_registry[self.socket.get_sid()] = modifier_registry[
                    self.socket.get_sid()
                ] + [cls.__name__]

            self._modifiers[cls.__name__] = {
                "cls": cls,
                "run_kwargs": run_kwargs,
                "public": public,
            }

        else:
            raise NotImplementedError(f"Variant {variant} is not implemented")

    def register_modifier(
        self,
        cls: t.Type[Extension],
        run_kwargs: dict | None = None,
        public: bool = False,
        timeout: float = 60,
        use_frozen: bool = True,
    ):
        """Register a modifier class.

        Attributes
        ----------
        cls : UpdateScene
            The modifier class to register.
        run_kwargs : dict, optional
            Keyword arguments to pass to the run method of the modifier class.
        public : bool, optional
            Whether to enable the modifier for ALL sessions of the ZnDraw client,
            or just the session for the given token.
        timeout : float, optional
            Timeout for the modifier to run in seconds. The Webclient
            will alert the user if the modifier takes longer than this time and
            release the modify button (no further changes are expected, but they
            can happen).
        use_frozen : bool, default=True
            Whether to use the ZnDrawFrozen class to run the modifier.
            The frozen class only allows provides cached data and
            e.g. access to other steps than the current one is not possible.
            If set to false, a full ZnDraw instance will be created for the modifier.
            This can have a performance impact and may lead to timeouts.

        """
        if run_kwargs is None:
            run_kwargs = {}

        # if use_frozen:
        #     cls = cls.frozen()

        self.register(
            cls=cls,
            run_kwargs=run_kwargs,
            public=public,
            variant=ExtensionType.MODIFIER,
        )


@tyex.deprecated("Use ZnDraw instead.")
@dataclasses.dataclass(kw_only=True)
class ZnDrawLocal(ZnDraw):
    """All the functionality is provided by the ZnDraw class."""



import eventlet

eventlet.monkey_patch()

import os

from engineio.payload import Payload

if max_decode_packets := os.environ.get("ENGINEIO_PAYLOAD_MAX_DECODE_PACKETS"):
    Payload.max_decode_packets = int(max_decode_packets)



import dataclasses
import importlib
import importlib.metadata
import importlib.util
import logging
import uuid

import znsocket
from flask import current_app, request, session
from flask_socketio import SocketIO, emit, join_room
from redis import Redis

from zndraw.tasks import (
    inspect_zntrack_node,
    load_zntrack_figures,
    load_zntrack_frames,
    run_room_copy,
    run_room_worker,
    run_scene_dependent_schema,
    run_schema,
    run_upload_file,
)

log = logging.getLogger(__name__)
__version__ = importlib.metadata.version("zndraw")


@dataclasses.dataclass
class DummyClient:
    """Dummy replacement for znsocket.Client."""

    sio: SocketIO
    refresh_callbacks: list = dataclasses.field(default_factory=list)


def init_socketio_events(io: SocketIO):
    @io.on("connect")
    def connect():
        emit("version", __version__)

    @io.on("shutdown")
    def shutdown():
        if "AUTH_TOKEN" not in current_app.config or session["authenticated"]:
            log.critical("Shutting down server")
            current_app.extensions["redis"].flushall()

            current_app.extensions["celery"].control.purge()
            current_app.extensions["celery"].control.broadcast("shutdown")

            io.stop()
        else:
            log.critical("Unauthenticated user tried to shut down the server.")

    @io.on("disconnect")
    def disconnect():
        try:
            room = str(session["token"])
        except KeyError:
            log.critical(f"disconnecting {request.sid}")
            return
        r = current_app.extensions["redis"]

        if "name" in session:
            log.critical(f"disconnecting (webclient) {request.sid} from room {room}")
            r.srem(f"room:{room}:webclients", session["name"])
            emit(
                "room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room
            )
        else:
            log.critical(f"disconnecting (pyclient) {request.sid}")

        for ref_token in [room, "default"]:
            modifier_registry = znsocket.Dict(
                r=r, key=f"registry:{ref_token}:modifier", repr_type="full"
            )

            modifier_schema = znsocket.Dict(
                r=r,
                key=f"schema:{ref_token}:modifier",
                repr_type="full",
                socket=DummyClient(sio=io),
            )

            # TODO: default room modifier
            for modifier in modifier_registry.pop(request.sid, []):
                for other in modifier_registry.values():
                    if modifier in other:
                        break
                else:
                    log.debug(f"Remove {modifier} from room {room}")
                    modifier_schema.pop(
                        modifier
                    )  # TODO this does not work well with public yet.

    @io.on("webclient:connect")
    def webclient_connect():
        try:
            token = session["token"]
        except KeyError:
            token = uuid.uuid4().hex[:8]
            session["token"] = token

        if "AUTH_TOKEN" not in current_app.config:
            session["authenticated"] = True
        else:
            if "authenticated" not in session:
                session["authenticated"] = False

        room = str(session["token"])
        join_room(room)  # rename token to room or room_token

        run_schema.delay(room)
        run_scene_dependent_schema.delay(room)

        session["name"] = uuid.uuid4().hex[:8]

        r = current_app.extensions["redis"]
        r.sadd(f"room:{room}:webclients", session["name"])

        # TODO: this is currently not used afaik
        emit("room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room)

        log.critical(f"connecting (webclient) {request.sid} to {room}")

        if "TUTORIAL" in current_app.config:
            emit("tutorial:url", current_app.config["TUTORIAL"])
        if "SIMGEN" in current_app.config:
            emit("showSiMGen", True)

        return {
            "name": session["name"],
            "room": room,
            "authenticated": session["authenticated"],
        }

    @io.on("join")  # rename pyclient:connect
    def join(data: dict):
        """
        Arguments:
            data: {"token": str, "auth_token": str}
        """
        # TODO: prohibit "token" to be "default"

        if "AUTH_TOKEN" not in current_app.config:
            session["authenticated"] = True
        else:
            session["authenticated"] = (
                data["auth_token"] == current_app.config["AUTH_TOKEN"]
            )
        token = data["token"]
        session["token"] = token
        room = str(session["token"])

        join_room(room)
        log.critical(f"connecting (pyclient) {request.sid} to {room}")
        # join_room(f"pyclients_{token}")

    @io.on("room:worker:run")
    def room_worker_run():
        """Start a worker to process all (available) queued tasks."""
        room = session.get("token")
        run_room_worker.delay(room)

    @io.on("schema:refresh")
    def schema_refresh():
        room = session.get("token")
        run_scene_dependent_schema.delay(room)

    @io.on("room:alert")
    def room_alert(msg: str):
        """Forward the alert message to every client in the room"""
        # TODO: identify the source client.
        room = session.get("token")
        emit("room:alert", msg, to=room)

    @io.on("room:upload:file")
    def room_upload_file(data: dict):
        room = session.get("token")
        run_upload_file.delay(room, data)

    @io.on("room:lock:set")
    def room_lock_set(locked: bool):
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        r.set(f"room:{room}:locked", str(locked))
        emit("room:lock:set", locked, to=room)

    @io.on("room:lock:get")
    def room_lock_get() -> bool:
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        locked = r.get(f"room:{room}:locked")
        return locked == "True"

    @io.on("room:token:get")
    def room_token_get() -> str:
        return session.get("token")

    @io.on("zntrack:available")
    def check_zntrack_available() -> bool:
        return importlib.util.find_spec("zntrack") is not None

    @io.on("zntrack:list-stages")
    def zntrack_list_stages(data: dict):
        try:
            import dvc.api

            fs = dvc.api.DVCFileSystem(url=data.get("remote"), rev=data.get("rev"))
            return [x.name for x in fs.repo.stage.collect() if hasattr(x, "name")]
        except Exception:
            return []

    @io.on("zntrack:inspect-stage")
    def zntrack_inspect_stage(data: dict):
        return inspect_zntrack_node(**data)

    @io.on("zntrack:load-frames")
    def zntrack_load_frames(data: dict):
        load_zntrack_frames.delay(room=session.get("token"), **data)

    @io.on("zntrack:load-figure")
    def zntrack_load_figure(data: dict):
        load_zntrack_figures.delay(room=session.get("token"), **data)

    @io.on("room:copy")
    def room_copy():
        room = session.get("token")
        run_room_copy.delay(room)



import importlib.metadata

from zndraw.base import Extension
from zndraw.draw import (
    Box,
    Circle,
    Cone,
    Custom2DShape,
    Cylinder,
    Dodecahedron,
    Ellipsoid,
    Icosahedron,
    Material,
    Octahedron,
    Plane,
    Rhomboid,
    Ring,
    Sphere,
    Tetrahedron,
    Torus,
    TorusKnot,
)
from zndraw.figure import Figure
from zndraw.zndraw import ZnDraw, ZnDrawLocal

__all__ = [
    "ZnDraw",
    "ZnDrawLocal",
    "Extension",
    "Plane",
    "Sphere",
    "Box",
    "Circle",
    "Cone",
    "Cylinder",
    "Dodecahedron",
    "Icosahedron",
    "Octahedron",
    "Ring",
    "Tetrahedron",
    "Torus",
    "TorusKnot",
    "Rhomboid",
    "Ellipsoid",
    "Material",
    "Custom2DShape",
    "Figure",
]

__version__ = importlib.metadata.version("zndraw")



import pathlib
import time

import redis
import znsocket.exceptions
from celery import Celery, Task
from flask import Flask
from flask_socketio import SocketIO

from zndraw.server import init_socketio_events, main_blueprint


def celery_init_app(app: Flask) -> Celery:
    class FlaskTask(Task):
        def __call__(self, *args: object, **kwargs: object) -> object:
            with app.app_context():
                return self.run(*args, **kwargs)

    celery_app = Celery(app.name, task_cls=FlaskTask)
    celery_app.config_from_object(app.config["CELERY"])
    celery_app.set_default()
    app.extensions["celery"] = celery_app
    return celery_app


def storage_init_app(app: Flask) -> None:
    if app.config["STORAGE"].startswith("redis"):
        app.extensions["redis"] = redis.Redis.from_url(
            app.config["STORAGE"], decode_responses=True
        )
    elif app.config["STORAGE"].startswith("znsocket"):
        for _ in range(100):  # try to connect to znsocket for 10 s
            # if we start znsocket via celery it will take some time to start
            try:
                app.extensions["redis"] = znsocket.Client.from_url(app.config["STORAGE"])
                break
            except ConnectionError:
                # wait for znsocket to start, if started together with the server
                time.sleep(0.1)
    else:
        raise ValueError(f"Unknown storage type: {app.config['STORAGE']}")


def create_app() -> Flask:
    app = Flask(__name__)
    app.config.update(SESSION_COOKIE_SAMESITE="None", SESSION_COOKIE_SECURE=True)
    app.config["SECRET_KEY"] = "secret!"
    # loads all FLASK_ prefixed environment variables into the app config
    app.config.from_prefixed_env()

    # TODO: this will not work without redis!!!!
    if app.config["STORAGE"].startswith("redis"):
        app.config.from_mapping(
            CELERY={
                "broker_url": app.config["STORAGE"],
                "result_backend": app.config["STORAGE"],
                "task_ignore_result": True,
            },
        )
    else:
        # nothing else supported, using filesystem storage
        data_folder = pathlib.Path("~/.zincware/zndraw/celery/out").expanduser()
        data_folder_processed = pathlib.Path(
            "~/.zincware/zndraw/celery/processed"
        ).expanduser()
        control_folder = pathlib.Path("~/.zincware/zndraw/celery/ctrl").expanduser()

        data_folder.mkdir(parents=True, exist_ok=True)
        data_folder_processed.mkdir(parents=True, exist_ok=True)
        control_folder.mkdir(parents=True, exist_ok=True)

        app.config.from_mapping(
            CELERY={
                "broker_url": "filesystem://",
                "result_backend": "cache",
                "cache_backend": "memory",
                "task_ignore_result": True,
                "broker_transport_options": {
                    "data_folder_in": data_folder.as_posix(),
                    "data_folder_out": data_folder.as_posix(),
                    "data_folder_processed": data_folder_processed.as_posix(),
                    "control_folder": control_folder.as_posix(),
                },
            },
        )

    # Initialize SocketIO
    message_queue = (
        app.config["CELERY"]["broker_url"]
        if app.config["STORAGE"].startswith("redis")
        else None
    )
    max_http_buffer_size = app.config.get("MAX_HTTP_BUFFER_SIZE")
    kwargs = {
        "message_queue": message_queue,
        "cors_allowed_origins": "*",
    }
    if "SOCKETIO_PING_TIMEOUT" in app.config:
        kwargs["ping_timeout"] = app.config["SOCKETIO_PING_TIMEOUT"]
    if max_http_buffer_size is not None:
        kwargs["max_http_buffer_size"] = int(max_http_buffer_size)

    socketio = SocketIO(app, **kwargs, logger=False, engineio_logger=False)

    # Initialize Celery
    celery_init_app(app)

    # Initialize storage
    storage_init_app(app)
    # we only need this server if we are using redis
    # otherwise a znsocket server will run anyhow
    if app.config["STORAGE"].startswith("redis"):
        from redis import Redis

        znsocket.attach_events(
            socketio.server,
            storage=Redis.from_url(app.config["STORAGE"], decode_responses=True),
        )
    else:
        znsocket.attach_events(
            socketio.server,
            storage=znsocket.Client.from_url(
                app.config["STORAGE"], decode_responses=True
            ),
        )

    # Register routes and socketio events
    app.register_blueprint(main_blueprint)
    init_socketio_events(socketio)

    # Add socketio to app extensions for easy access
    app.extensions["socketio"] = socketio

    return app



import logging
import uuid
from io import BytesIO, StringIO

import ase.io
from flask import (
    Blueprint,
    current_app,
    redirect,
    request,
    send_file,
    send_from_directory,
    session,
)

main = Blueprint("main", __name__)

log = logging.getLogger(__name__)


@main.route("/")
def index():
    """Render the main ZnDraw page."""
    try:
        token = session["token"]
    except KeyError:
        token = uuid.uuid4().hex[:8]
        session["token"] = token

    request_args = request.args.to_dict()
    if len(request_args) > 0:
        from zndraw import ZnDraw

        vis = ZnDraw(
            r=current_app.extensions["redis"],
            url=current_app.config["SERVER_URL"],
            token=session.get("token"),
        )
        if "step" in request_args:
            vis.step = int(request_args["step"])
        if "selection" in request_args:
            if request_args["selection"] == "null":
                vis.selection = []
            else:
                vis.selection = [int(i) for i in request_args["selection"].split(",")]

    if "APPLICATION_ROOT" in current_app.config:
        return redirect(f"{current_app.config['APPLICATION_ROOT']}token/{token}")
    return redirect(f"/token/{token}")


@main.route("/<path:filename>")
def main_files(filename):
    return send_from_directory("templates/", filename)


@main.route("/assets/<path:filename>")
def assets(filename):
    return send_from_directory("templates/assets", filename)


@main.route("/token/<token>")
def token(token):
    if token == "default":
        return "Invalid token", 403
    session["token"] = token
    return send_from_directory("templates", "index.html")


@main.route("/reset")
def reset():
    session["token"] = uuid.uuid4().hex[:8]  # TODO: how should reset work locally?
    if "APPLICATION_ROOT" in current_app.config:
        return redirect(
            f"{current_app.config['APPLICATION_ROOT']}token/{session['token']}"
        )
    return redirect(f"/token/{session['token']}")


@main.route("/exit")
@main.route("/exit/<token>")
def exit_route(token: str | None = None):
    """Exit the session."""
    log.critical("Server shutting down...")
    if not session.get("authenticated", False) and token != current_app.config.get(
        "AUTH_TOKEN"
    ):
        return "Invalid auth token", 403

    current_app.extensions["socketio"].stop()
    return "Server shutting down..."


@main.route("/login/<auth_token>")
def login_route(auth_token: str | None = None):
    """Create an authenticated session."""
    session["authenticated"] = auth_token == current_app.config.get("AUTH_TOKEN", "NONE")
    if session["authenticated"]:
        if "APPLICATION_ROOT" in current_app.config:
            return redirect(f"{current_app.config['APPLICATION_ROOT']}")
        return redirect("/")
    return "Invalid auth token", 403


@main.route("/logout")
def logout_route():
    if not session.get("authenticated", False):
        return "Can only log out, if you logged in before.", 403
    session["authenticated"] = False
    if "APPLICATION_ROOT" in current_app.config:
        return redirect(f"{current_app.config['APPLICATION_ROOT']}")
    return redirect("/")


@main.route("/upload", methods=["POST"])
def upload():
    """Upload a file to the server."""
    from zndraw import ZnDrawLocal

    file = request.files["file"]
    token = session.get("token")

    if not token:
        return "Unauthorized", 401

    try:
        # Extract the file format from the filename
        file_format = file.filename.split(".")[-1]
        file_content = file.read()

        stream = StringIO(file_content.decode("utf-8"))

        vis = ZnDrawLocal(
            r=current_app.extensions["redis"],
            url=current_app.config["SERVER_URL"],
            token=token,
        )
        structures = list(ase.io.iread(stream, format=file_format))
        vis.extend(structures)
        vis.socket.disconnect()

        return "File uploaded", 200

    except Exception as e:
        log.error(f"Error uploading file: {e}")
        return str(e), 500


@main.route("/download", methods=["GET"])
def download():
    """Download a file to the client."""
    from zndraw import ZnDrawLocal

    token = session.get("token")

    file = StringIO()
    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=token,
    )
    try:
        for atoms in vis:
            ase.io.write(file, atoms, format="xyz", append=True)
    except Exception as e:
        log.error(f"Error downloading file: {e}")

    # convert StringIO to BytesIO
    file = BytesIO(file.getvalue().encode("utf-8"))
    try:
        return send_file(file, as_attachment=True, download_name="trajectory.xyz")
    except Exception as e:
        log.error(f"Error downloading file: {e}")
        return str(e), 500
